@@ -1,6 +1,6 @@
 # CSV -- module for generating/parsing CSV data.
   
-# $Id: csv.rb,v 1.10 2002/04/13 10:33:53 nahi Exp $
+# $Id: csv.rb,v 1.16 2002/12/12 23:01:10 nahi Exp $
   
 # This module is copyrighted free software by NAKAMURA, Hiroshi.
 # You can redistribute it and/or modify it under the same term as Ruby.
@@ -253,6 +253,7 @@ public
   #   end
   #
   class Reader
+    include Enumerable
   public
 
     # SYNOPSIS
@@ -409,6 +410,9 @@ public
       @colSep = colSep
       @dev = string
       @idx = 0
+      if @dev[0, 3] == "\xef\xbb\xbf"
+	@idx += 3
+      end
     end
 
   private
@@ -454,6 +458,9 @@ public
       @colSep = colSep
       @dev = CSV::IOBuf.new( @io )
       @idx = 0
+      if @dev[0] == 0xef and @dev[1] == 0xbb and @dev[2] == 0xbf
+	@idx += 3
+      end
       @closeOnTerminate = false
     end
 
@@ -694,15 +701,14 @@ public
   #   not, use CSV.parseLine instead of this method.
   #   
   def CSV.parse( src, colSep = ?, )
-    nextIdx = 0
+    idx = 0
     resType = :DT_COLSEP
     cells = Row.new
     begin
       while ( resType.equal?( :DT_COLSEP ))
 	aCell = Cell.new
-	resType, nextIdx = parseBody( src, nextIdx, aCell, colSep )
+	resType, idx = parseBody( src, idx, aCell, colSep )
 	cells.push( aCell.isNull ? nil : aCell.data )
-	resType, nextIdx = parseSeparator( src, nextIdx, colSep )
       end
     rescue IllegalFormatError
       return Row.new
@@ -781,20 +787,18 @@ public
   #   end while nofCells > 0
   #   
   def CSV.parseLine( src, idx, outDev, colSep = ?, )
+    idxBack = idx
     nofCell = 0
-    nextIdx = idx
     resType = :DT_COLSEP
     begin
-      while ( resType.equal?( :DT_COLSEP ))
+      while ( !resType.equal?( :DT_ROWSEP ))
 	aCell = Cell.new
-	resType, nextIdx = parseBody( src, nextIdx, aCell, colSep )
-	resType, nextIdx = parseSeparator( src, nextIdx, colSep )
+	resType, idx = parseBody( src, idx, aCell, colSep )
 	if resType.equal?( :DT_EOS )
-	  if nofCell > 0 || !aCell.isNull
-	    resType = :DT_ROWSEP
-	  else
-	    break
+	  if idx == idxBack #(( nofCell == 0 ) && ( aCell.isNull ))
+	    return 0, 0
 	  end
+	  resType = :DT_ROWSEP
 	end
 	nofCell += 1
 	outDev << aCell
@@ -802,10 +806,7 @@ public
     rescue IllegalFormatError
       return 0, 0
     end
-    unless resType.equal?( :DT_ROWSEP )
-      return 0, 0
-    end
-    return nofCell, nextIdx
+    return nofCell, idx
   end
   
   # SYNOPSIS
@@ -875,25 +876,22 @@ public
   private
 
     def parseBody( src, idx, aCell, colSep )
-      nextIdx = idx
-      quoted = false
-      cr = false
       aCell.isNull = false
       state = :ST_START
-      while (( c = src[ nextIdx ] ) && !state.equal?( :ST_END ))
+      quoted = false
+      cr = false
+      c = nil
+      while ( c = src[ idx ] )
+	idx += 1
+	resultState = :DT_UNKNOWN
 	if ( c == colSep )
-	  if state.equal?( :ST_START )
-	    aCell.isNull = true
-	    nextIdx -= 1
-	    state = :ST_END
-	  elsif state.equal?( :ST_DATA )
+	  if state.equal?( :ST_DATA )
 	    if cr
-	      aCell.data << "\r"
-	      cr = false
+	      raise IllegalFormatError.new
 	    end
 	    if ( !quoted )
-	      nextIdx -= 1
 	      state = :ST_END
+	      resultState = :DT_COLSEP
 	    else
 	      aCell.data << c.chr
 	    end
@@ -901,19 +899,17 @@ public
 	    if cr
 	      raise IllegalFormatError.new
 	    end
-	    nextIdx -= 1
 	    state = :ST_END
-	  else
-	    raise RuntimeError.new( 'Must not reach here.' )
+	    resultState = :DT_COLSEP
+	  else	# :ST_START
+	    aCell.isNull = true
+	    state = :ST_END
+	    resultState = :DT_COLSEP
 	  end
 	elsif ( c == ?" )		# " for vim syntax hilighting.
-	  if state.equal?( :ST_START )
-	    quoted = true
-	    state = :ST_DATA
-	  elsif state.equal?( :ST_DATA )
+	  if state.equal?( :ST_DATA )
 	    if cr
-	      aCell.data << "\r"
-	      cr = false
+	      raise IllegalFormatError.new
 	    end
 	    if quoted
 	      quoted = false
@@ -922,112 +918,82 @@ public
 	      raise IllegalFormatError.new
 	    end
 	  elsif state.equal?( :ST_QUOTE )
+	    aCell.data << c.chr
+	    quoted = true
+	    state = :ST_DATA
+	  else	# :ST_START
+	    quoted = true
+	    state = :ST_DATA
+	  end
+	elsif ( c == ?\r )
+	  if cr
+	    raise IllegalFormatError.new
+	  end
+	  if quoted
+	    aCell.data << c.chr
+	    state = :ST_DATA
+	  else
+	    cr = true
+	  end
+	elsif ( c == ?\n )
+	  if state.equal?( :ST_DATA )
+	    if cr
+	      state = :ST_END
+	      resultState = :DT_ROWSEP
+	      cr = false
+	    else
+	      if quoted
+		aCell.data << c.chr
+		state = :ST_DATA
+	      else
+		state = :ST_END
+		resultState = :DT_ROWSEP
+	      end
+	    end
+	  elsif state.equal?( :ST_QUOTE )
+	    state = :ST_END
+	    resultState = :DT_ROWSEP
+	    if cr
+	      cr = false
+	    end
+	  else	# :ST_START
+	    aCell.isNull = true
+	    state = :ST_END
+    	    resultState = :DT_ROWSEP
+	  end
+	else
+	  if state.equal?( :ST_DATA ) || state.equal?( :ST_START )
 	    if cr
 	      raise IllegalFormatError.new
 	    end
 	    aCell.data << c.chr
-	    quoted = true
 	    state = :ST_DATA
-	  else
-	    raise RuntimeError.new( 'Must not reach here.' )
-	  end
-	elsif ( c == ?\r )
-	  if state.equal?( :ST_START )
-	    state = :ST_DATA
-	  end
-	  cr = true
-	elsif ( c == ?\n )
-	  if state.equal?( :ST_START )
-	    aCell.data << c.chr
-	  state = :ST_DATA
-	  elsif state.equal?( :ST_DATA )
-	    if cr
-	      if quoted
-		aCell.data << "\r" << c.chr
-		state = :ST_DATA
-	      else
-		if aCell.data.empty?
-		  aCell.isNull = true
-		end
-		nextIdx -= 2
-		state = :ST_END
-	      end
-	      cr = false
-	    else
-	      aCell.data << c.chr
-	      state = :ST_DATA
-	    end
-	  elsif state.equal?( :ST_QUOTE )
-	    if cr
-	      nextIdx -= 2
-	      state = :ST_END
-	      cr = false
-	    else
-	      raise RuntimeError.new( 'Must not reach here.' )
-	    end
-	  else
-	    raise RuntimeError.new( 'Must not reach here.' )
-	  end
-	else
-	  if state.equal?( :ST_START ) || state.equal?( :ST_DATA )
-	    if cr
-	      aCell.data << "\r"
-	      cr = false
-	    end
-	    aCell.data << c.chr
-	    state = :ST_DATA
-	  elsif state.equal?( :ST_QUOTE )
+	  else	# :ST_QUOTE
 	    raise IllegalFormatError.new
-	  else
-	    raise RuntimeError.new( 'Must not reach here.' )
 	  end
 	end
-	nextIdx += 1
+	if state.equal?( :ST_END )
+	  return resultState, idx;
+	end
       end
       if state.equal?( :ST_START )
 	aCell.isNull = true
-	return :DT_BODY, nextIdx
-      elsif state.equal?( :ST_END )
-	return :DT_BODY, nextIdx
-      elsif ( src[ nextIdx ].nil? )
-	if quoted
-	  raise IllegalFormatError.new
-	end
-	if cr
-	  aCell.data << "\r"
-	  cr = false
-	end
-	return :DT_BODY, nextIdx
-      else
-	raise RuntimeError.new( 'Must not reach here.' )
+      elsif state.equal?( :ST_QUOTE )
+	true	# dummy for coverate; only a data
+      elsif quoted
+       	raise IllegalFormatError.new
+      elsif cr
+       	raise IllegalFormatError.new
       end
+      return :DT_EOS, idx
     end
   
-    def parseSeparator( src, idx, colSep )
-      nextIdx = idx
-      resType = :DT_FAIL
-      if src[ nextIdx ].nil?
-	resType = :DT_EOS
-      elsif ( src[ nextIdx ] == colSep )
-	nextIdx += 1
-	resType = :DT_COLSEP
-      elsif (( src[ nextIdx ] == ?\r ) && ( src[ nextIdx + 1 ] == ?\n ))
-	nextIdx += 2
-	resType = :DT_ROWSEP
-      else
-	# Thinking about the spec of parseBody, it must not reach here.
-	raise RuntimeError.new( 'Must not reach here.' )
-      end
-      return resType, nextIdx
-    end
-    
     def createBody( cellData, outDev, colSep )
       addData = cellData.data.dup
       if ( !cellData.isNull )
        	if ( addData.gsub!( '"', '""' ) ||
 	    addData.include?( colSep ) ||
-	    addData.include?( ?" ) ||	# "
-  	    ( addData =~ /\r\n/ ) || ( cellData.data.empty? ))
+  	    ( /[\r\n]/ =~ addData ) || ( cellData.data.empty? ))
   	  outDev << '"' << addData << '"'
    	else
   	  outDev << addData
@@ -1041,8 +1007,6 @@ public
        	outDev << colSep.chr
       when :DT_ROWSEP
 	outDev << "\r\n"
-      else
-	raise RuntimeError.new( 'Must not reach here.' )
       end
     end
   end
@@ -1161,8 +1125,7 @@ public
 	return @bufList[ myBuf ][ myOffset + nextIdx ]	# Fixnum of char code.
       elsif ( myOffset + nextIdx + n - 1 < BufSize )
 	return @bufList[ myBuf ][ myOffset + nextIdx, n ]	# String.
-      else
-	# should do loop insted of (tail) recursive call...
+      else # should do loop insted of (tail) recursive call...
 	res = @bufList[ myBuf ][ myOffset + nextIdx..-1 ]
 	addSize = BufSize - ( myOffset + nextIdx )
 	idx += addSize
